@@ -7,10 +7,16 @@ except ImportError:
 from functools import wraps
 import hashlib
 
+import django
 from django.core.exceptions import ImproperlyConfigured
 from django.db.models import Manager, Model
 from django.db.models.query import QuerySet, ValuesQuerySet, ValuesListQuerySet, DateQuerySet
-from django.db.models.signals import post_save, post_delete, m2m_changed
+from django.db.models.signals import pre_save, post_save, post_delete, m2m_changed
+
+try:
+    from django.db.models.query import MAX_GET_RESULTS
+except ImportError:
+    MAX_GET_RESULTS = None
 
 from cacheops.conf import model_profile, redis_client, handle_connection_failure
 from cacheops.utils import monkey_mix, dnf, conj_scheme, get_model_name, non_proxy, stamp_fields
@@ -132,16 +138,13 @@ def _stringify_query():
           any significant optimization will most likely require a major
           refactor of sql.Query class, which is a substantial part of ORM.
     """
-    # import django
-    # print django.VERSION
-
-
     import simplejson as json
     from datetime import datetime, date
     from django.db.models.expressions import ExpressionNode, F
     from django.db.models.fields import Field
-    from django.db.models.fields.related import ManyToOneRel
-    from django.db.models.sql.where import Constraint, WhereNode, ExtraWhere
+    from django.db.models.fields.related import ManyToOneRel, OneToOneRel
+    from django.db.models.sql.where import Constraint, WhereNode, ExtraWhere, \
+                                           EverythingNode, NothingNode
     from django.db.models.sql import Query
     from django.db.models.sql.aggregates import Aggregate
     from django.db.models.sql.datastructures import Date
@@ -149,6 +152,14 @@ def _stringify_query():
 
     attrs = {}
 
+    # A new things in Django 1.6
+    try:
+        from django.db.models.sql.where import EmptyWhere, SubqueryConstraint
+        attrs[EmptyWhere] = ()
+        attrs[SubqueryConstraint] = ('alias', 'columns', 'targets', 'query_object')
+    except ImportError:
+        pass
+
     # RawValue removed in Django 1.7
     try:
         from django.db.models.sql.datastructures import RawValue
@@ -163,21 +174,19 @@ def _stringify_query():
     attrs[Aggregate] = ('source', 'is_summary', 'col', 'extra')
     attrs[Date] = ('col', 'lookup_type')
     attrs[F] = ('name',)
-    attrs[ManyToOneRel] = ('field',)
+    attrs[ManyToOneRel] = attrs[OneToOneRel] = ('field',)
+    attrs[EverythingNode] = attrs[NothingNode] = ()
 
     q = Query(None)
     q_keys = q.__dict__.keys()
     q_ignored = ['join_map', 'dupe_avoidance', '_extra_select_cache', '_aggregate_select_cache',
                  'used_aliases']
     attrs[Query] = tuple(sorted( set(q_keys) - set(q_ignored) ))
-    # print 'QQQ>', attrs[Query]
 
     for k, v in attrs.items():
         attrs[k] = map(intern, v)
 
     def encode_object(obj):
-        # if not isinstance(obj, type):
-        #     print '>>>>>', obj, getattr(obj, '__dict__', None)
         if isinstance(obj, set):
             return sorted(obj)
         elif isinstance(obj, type):
@@ -339,6 +348,7 @@ class QuerySetMixin(object):
         return clone
 
     def iterator(self):
+        # TODO: do not cache empty queries in Django 1.6
         superiter = self._no_monkey.iterator
         cache_this = self._cacheprofile and 'fetch' in self._cacheops
 
@@ -367,7 +377,8 @@ class QuerySetMixin(object):
     def count(self):
         # Optmization borrowed from overriden method:
         # if queryset cache is already filled just return its len
-        if self._result_cache is not None and not self._iter:
+        # NOTE: there is no self._iter in Django 1.6+, so we use getattr() for compatibility
+        if self._result_cache is not None and not getattr(self, '_iter', None):
             return len(self._result_cache)
         return cached_method(op='count')(self._no_monkey.count)(self)
 
@@ -418,7 +429,7 @@ class QuerySetMixin(object):
                and query_dnf[0][0][0] == self.model._meta.pk.name:
                 result = len(self.nocache()) > 0
                 if result:
-                    _old_objs[get_model_name(self.model)][query_dnf[0][0][1]] = self._result_cache[0]
+                    _old_objs[self.model][query_dnf[0][0][1]] = self._result_cache[0]
                 return result
         return self._no_monkey.exists(self)
 
@@ -426,11 +437,15 @@ class QuerySetMixin(object):
 class ManagerMixin(object):
     def _install_cacheops(self, cls):
         cls._cacheprofile = model_profile(cls)
-        if cls._cacheprofile is not None and get_model_name(cls) not in _old_objs:
+        if cls._cacheprofile is not None and cls not in _old_objs:
             # Set up signals
+            if not getattr(cls._meta, 'select_on_save', True):
+                # Django 1.6+ doesn't make select on save by default,
+                # which we use to fetch old state, so we fetch it ourselves in pre_save handler
+                pre_save.connect(self._pre_save, sender=cls)
             post_save.connect(self._post_save, sender=cls)
             post_delete.connect(self._post_delete, sender=cls)
-            _old_objs[get_model_name(cls)] = {}
+            _old_objs[cls] = {}
 
             # Install auto-created models as their module attributes to make them picklable
             module = sys.modules[cls.__module__]
@@ -441,11 +456,17 @@ class ManagerMixin(object):
         self._no_monkey.contribute_to_class(self, cls, name)
         self._install_cacheops(cls)
 
+    def _pre_save(self, sender, instance, **kwargs):
+        try:
+            _old_objs[sender][instance.pk] = sender.objects.get(pk=instance.pk)
+        except sender.DoesNotExist:
+            pass
+
     def _post_save(self, sender, instance, **kwargs):
         """
         Invokes invalidations for both old and new versions of saved object
         """
-        old = _old_objs[get_model_name(instance.__class__)].pop(instance.pk, None)
+        old = _old_objs[sender].pop(instance.pk, None)
         if old:
             invalidate_obj(old)
         invalidate_obj(instance)
@@ -476,7 +497,9 @@ class ManagerMixin(object):
             filter_key = key[:-3] if key.endswith('_id') else key
 
             cond = {filter_key: getattr(instance, key)}
-            qs = instance.__class__.objects.inplace().filter(**cond).order_by()
+            qs = sender.objects.inplace().filter(**cond).order_by()
+            if MAX_GET_RESULTS:
+                qs = qs[:MAX_GET_RESULTS + 1]
             qs._cache_results(qs._cache_key(), [instance])
 
             # Reverting stripped attributes
@@ -490,17 +513,21 @@ class ManagerMixin(object):
         #       before deletion (why anyone will do that?)
         invalidate_obj(instance)
 
+    # Django 1.5- compatability
+    if django.VERSION < (1, 6):
+        get_queryset = Manager.get_query_set
+
     def inplace(self):
-        return self.get_query_set().inplace()
+        return self.get_queryset().inplace()
 
     def get(self, *args, **kwargs):
-        return self.get_query_set().inplace().get(*args, **kwargs)
+        return self.get_queryset().inplace().get(*args, **kwargs)
 
     def cache(self, *args, **kwargs):
-        return self.get_query_set().cache(*args, **kwargs)
+        return self.get_queryset().cache(*args, **kwargs)
 
     def nocache(self, *args, **kwargs):
-        return self.get_query_set().nocache(*args, **kwargs)
+        return self.get_queryset().nocache(*args, **kwargs)
 
 
 def invalidate_m2m(sender=None, instance=None, model=None, action=None, pk_set=None, **kwargs):
